"""
Copyright 2022 nMachine.io

Requirements:
- e2e cluster, configuration from KUBECONFIG or ~/.kube/config
- applied configuration from k8kat/e2e_tests/cluster-e2e.yaml

"""

import unittest

from k8kat import Ingress


class TestIngress(unittest.TestCase):
    """Set of e2e tests for Ingress resource."""

    def test_list_ingress_from_e2e_namespace(self) -> None:
        """Tests listing Ingresses from 'e2e' namespace"""
        ingresses = Ingress.list(ns="e2e")
        self.assertEqual(len(ingresses), 1)
        self.assertEqual(ingresses[0].name, "example-ingress")

    def test_find_ingress_with_not_found_error(self) -> None:
        """Tests finding not-existing Ingress"""
        ingress = Ingress.find("ingress-does-not-exist", ns="e2e")
        self.assertIsNone(ingress)

    def test_find_ingress(self) -> None:
        """Tests finding ingress"""
        sec = Ingress.find("example-ingress", ns="e2e")
        assert sec

        self.assertEqual(sec.name, "example-ingress")
        self.assertEqual(sec.labels, {})
        self.assertEqual(
            sec.flat_rules(),
            [
                {
                    "host": "localhost",
                    "service": "foo-service",
                    "port": 5678,
                    "path": "/foo",
                },
                {
                    "host": "localhost",
                    "service": "bar-service",
                    "port": 5678,
                    "path": "/bar",
                },
            ],
        )
        self.assertEqual(
            sec.basic_rules(),
            {
                "localhost": [
                    {"path": "/foo", "port": 5678, "service": "foo-service"},
                    {"path": "/bar", "port": 5678, "service": "bar-service"},
                ]
            },
        )



"""This module provides classes required to communicate
with Kubernetes clusters.

There two classes:

    Configuration() which provides parameters (adresses,
        credentials etc.) required to connect to a cluster.

    Broker() is a class which exposes endpoints to connect
        to get/set data in K8s clusters.


Both classes are required by K8Kat's resources. The instance
of broker (singleton) will be created if you don't pass
your own instance of the broker.

Example:

    config = Configuration(kubeconfig_path='/my/kube/config')
    broker = Broker(config)
    ...
    pods = broker.core_v1_api.list_pod_for_all_namespaces()


Copyright 2022 nMachine.io
"""

from __future__ import annotations

from typing import Optional

from kubernetes import client, config, dynamic  # type: ignore

# pylint: disable=unused-import
from kubernetes.config.config_exception import ConfigException  # type: ignore


class Configuration:  # pylint: disable=too-few-public-methods
    """Configuration for Broker."""

    def __init__(
        self,
        token: Optional[str] = None,
        api_server: Optional[str] = None,
        ca_cert: Optional[str] = None,
        verify_ssl: Optional[bool] = None,
        kubeconfig_path: Optional[str] = None,
    ) -> None:
        """Configuration for the class Broker. Use arguments: token,
        api_server, ca_cert, verify_ssl to authorize using the
        bearer token.

        Optionally it tries to load configuration from kubeconfig
        file from standard locations, defined in environment variable
        or passed by kubeconfig_path.

        It also handles the situation when it runs in "in-cluster" mode.

        Args:
            token(str): Bearer token
            api_server(str): address to API server
            ca_cert(str): cluster's certificate authority
            verify_ssl(bool): option to ignore server cert, ignored if ca_cert is passed

            kubeconfig_path(str): path to Kube Config file

        """
        self._kubernetes_config = client.Configuration()

        if api_server is not None:
            self._kubernetes_config.host = api_server
            self._kubernetes_config.api_key = {"authorization": f"Bearer {token}"}
            if ca_cert:
                self._kubernetes_config.verify_ssl = True
                self._kubernetes_config.ssl_ca_cert = ca_cert
            else:
                if verify_ssl is None:
                    verify_ssl = False
                self._kubernetes_config.verify_ssl = verify_ssl

        elif kubeconfig_path:
            config.load_kube_config(  # TODO: config.load_config()
                config_file=kubeconfig_path,
                client_configuration=self.kubernetes_config,
            )
        else:
            config.load_config(client_configuration=self.kubernetes_config)

    @property
    def kubernetes_config(self) -> client.Configuration:
        """Returns low-level client.Configuration from the kubernetes library."""
        return self._kubernetes_config


class Broker:  # pylint: disable=too-few-public-methods
    """Class privides endpoints, alrady configured to query K8s clusters.
    If you want to work with many clusters in the same application
    you will need to create Brokers manually. In simpler cases default
    Broker, automatically created should be enough.

    Class attributes:
        default_instance : singleton object if created

    Attributes:

        core_v1_api : raw K8s interface to get/set core resources
        custom_api : raw K8s interface to get/set custom resources

    """

    default_instance = None

    def __init__(self, configuration: Optional[Configuration] = None):
        """Broker constructor, the object can be created with custom configuration.
        If it's not provided it uses a default configuration.

        Args:
            configuration (Configuration): configuration for broker
        """

        if configuration is None:
            configuration = Configuration()
        self._configuration = configuration
        self._api_client = client.ApiClient(configuration.kubernetes_config)

        self.core_v1_api = client.CoreV1Api(self._api_client)
        self.custom_api = client.CustomObjectsApi(self._api_client)
        self.networking_v1_api = client.NetworkingV1Api(self._api_client)

        self.dynamic_api = dynamic.DynamicClient(
            client.api_client.ApiClient(
                configuration=self._configuration.kubernetes_config
            )
        )

    @classmethod
    def get_default_instance(cls) -> Broker:
        """Returns default instance of Broker if created automatically"""
        if cls.default_instance is None:
            cls.default_instance = Broker()
        return cls.default_instance



"""It provides BaseResource class.

Copyright 2022 nMachine.io
"""

import time
from datetime import datetime
from typing import Any, Callable, Dict, List, Optional, Type, TypeVar

from kubernetes.client import ApiException  # type: ignore

from k8kat.broker import Broker
from k8kat.res.selectors import convert_selectors
from k8kat.utils import classproperty

KR = TypeVar("KR", bound="BaseResource")


class BaseResource:
    """Base class for other resources like Pod, ConfigMap etc."""

    def __init__(self, raw: Any, broker: Broker):
        """Raw object and Broker are required to construct base resource.

        Args:
            raw: object returned from Kubernetes library
            broker: source Broker which return this object, will be used
                    to patch, reload or delete the object

        """

        self.raw = raw  # raw object from Kubernetes library.
        self.broker = broker  # Broker which returns this resource

    def as_dict(self) -> Dict:
        """Returns objects attributes as dict"""
        return self.raw.to_dict()

    @property
    def uid(self) -> str:
        """Returns UID from metadata"""
        return self.raw.metadata.uid

    @classproperty
    def res_name_plural(self) -> str:
        """Returns plural name of the resource"""
        kind_lower = str(self.kind).lower()
        return f"{kind_lower}s"

    @classproperty
    def kind_aliases(self) -> List[str]:
        """Returns aliases for this type of resource"""
        return [self.kind, str(self.kind).lower(), self.res_name_plural]

    @property
    def name(self) -> str:
        """Returns name of the object"""
        return self.raw.metadata.name

    @property
    def namespace(self) -> str:
        """Returns name of the namespace"""
        return self.raw.metadata.namespace

    @property
    def ns(self) -> str:
        """Returns name of the namespace"""
        return self.namespace

    def sig(self, with_ns: Optional[bool] = False) -> str:
        """Returns name in format namespace:kind:name

        Returns:
            string as extended name"""
        ns_less = f"{self.kind}:{self.name}"
        return f"{self.ns}:{ns_less}" if with_ns else ns_less

    @property
    def metrics_component(self) -> Optional[Dict]:
        """Returns object's run time metrics."""
        return None

    @property
    def labels(self) -> Dict[str, str]:
        """Returns dict of labels"""
        return self.raw.metadata.labels or {}

    @property
    def annotations(self) -> Dict[str, str]:
        """Returns dict of annotations"""
        return self.raw.metadata.annotations or {}

    def created_at(self) -> Optional[datetime]:
        """Returns timestamp when object has been created

        Returns:
            object creation datetime"""
        value: datetime = self.raw.metadata.creation_timestamp
        return value.replace(tzinfo=None) if value else None

    def seconds_existed(self) -> int:
        """Returns how long the object lives (in second)

        Returns:
            live time in secods"""
        created_ts = self.created_at()
        if created_ts:
            return (datetime.utcnow() - created_ts).seconds
        return 1_000_000_000

    def wait_until(
        self, predicate: Callable, max_time_sec: Optional[float] = None
    ) -> bool:
        """Waits until pedicated() return true, reloads the object
        periodically.

        Args:
            predicate : function to check
            max_time_sec : timeout

        Returns:
            True if condition met, False otherwise
        """
        start_time = time.time()
        condition_met = False
        for _ in range(0, 1_000):
            if predicate():
                condition_met = True
                break
            if max_time_sec and time.time() - start_time > max_time_sec:
                return False
            time.sleep(1)
            self.reload()
        return condition_met

    def reload(self) -> Any:
        """Reloads object state by querying Kubernetes API.

        Returns:
            self refreshed or None if object disappeared"""
        self.raw = self.find_raw(self.name, self.ns, self.broker)
        return self if self.raw else None

    def patch(self, modifier: Optional[Callable] = None) -> Optional[KR]:
        """Patches object using modifier or self.raw object

        Args:
            modifier : function which modifies the object

        Returns:
            reloaded object

        """

        if modifier is not None:
            self._enter_patch_loop(modifier)
        else:
            self.ns_agnostic_call(
                self.k8s_patch_method, body=self.raw, broker=self.broker
            )
        return self.reload()

    def _enter_patch_loop(self, modification_delegate: Callable) -> None:
        failed_attempts = 0
        while True:
            try:
                modification_delegate(self.raw)
                self.ns_agnostic_call(
                    self.k8s_patch_method, body=self.raw, broker=self.broker
                )
                return
            except ApiException:
                if failed_attempts >= 5:
                    raise
                failed_attempts += 1
                time.sleep(1)
                self.reload()

    def ns_agnostic_call(self, impl: Callable, **kwargs: Any) -> Any:
        """Calls method impl with name and with the namespace
        if the resource is namespaced.

        Args:
            impl: method to call

        Returns:
            values from impl
        """
        if self.is_namespaced():
            return impl(name=self.name, namespace=self.ns, **kwargs)
        return impl(name=self.name, **kwargs)

    def delete(self, wait_until_gone: Optional[bool] = False) -> None:
        """Deletes the object

        Args:
            wait_until_gone: True if you want to wait for
                             the object to disappear.

        """
        self.ns_agnostic_call(self.k8s_delete_method, broker=self.broker)

        if wait_until_gone:
            while self.reload():
                time.sleep(0.5)

    @classproperty
    def kind(self) -> str:
        """Name of the resource"""
        raise NotImplementedError

    @classmethod
    def k8s_list_method(cls, broker: Broker, **kwargs: Any) -> Any:
        """Returns list of raw objects using K8s API"""
        raise NotImplementedError

    @classmethod
    def k8s_read_method(cls, broker: Broker, **kwargs: Any) -> Any:
        """Returns only one raw object using K8s API"""
        raise NotImplementedError

    @classmethod
    def k8s_patch_method(cls, broker: Broker, **kwargs: Any) -> Any:
        """Patches the raw object using K8s API"""
        raise NotImplementedError

    @classmethod
    def k8s_delete_method(cls, broker: Broker, **kwargs: Any) -> None:
        """Deletes object using K8s API"""
        raise NotImplementedError

    @classmethod
    def k8s_apply_method(cls, broker: Broker, **kwargs: Any) -> None:
        """Equivalent to kubectl apply -f using K8s API"""
        raise NotImplementedError

    @classmethod
    def is_namespaced(cls) -> bool:
        """True if resource is namespaced"""
        return True

    @classmethod
    def list(
        cls: Type[KR],
        ns: Optional[str] = None,
        broker: Optional[Broker] = None,
        **kwargs: Any,
    ) -> List[KR]:
        """Returns list of objects.

        Args:
            ns: namespace
            broker: Broker to query K8s API

            field_selector: str, K8s string with expression to include/exclude
                            resources
            fields: List[Tuple(str, str)], list of pair (field name, field value) which
                    have to match in returned resources
            not_fields: List[Tuple(str, str)], list of pair (field name, field value) which
                    have to be excluded from resources

            label_selector: str, K8s string with expression to include/exclude
                            resources
            labels: List[Tuple(str, str)], list of pair (label name, label value) which
                    have to match in returned resources
            not_labels: List[Tuple(str, str)], list of pair (label name, label value) which
                    have to be excluded from resources

            limit: int, limit returned resources

        Returns:
            list of objects
        """
        if broker is None:
            broker = Broker.get_default_instance()
        convert_selectors(kwargs)
        k8s_items = cls.k8s_list_method(broker, ns=ns, **kwargs)
        return [cls(raw, broker=broker) for raw in k8s_items.items]

    @classmethod
    def find_raw(
        cls, name: str, ns: Optional[str] = None, broker: Optional[Broker] = None
    ) -> Any:
        """Finds raw object by name.

        Args:
            name: name of the object
            ns: namespace
            broker: Broker to query K8s API

        Returns:
            return raw k8s object or None if not found
        """
        try:
            if broker is None:
                broker = Broker.get_default_instance()
            fun: Callable = cls.k8s_read_method
            is_ns: bool = cls.is_namespaced()
            return (
                fun(name=name, ns=ns, broker=broker)
                if is_ns
                else fun(name=name, broker=broker)
            )
        except ApiException as ex:
            if ex.status != 404:
                raise
        return None

    @classmethod
    def find(
        cls: Type[KR],
        name: str,
        ns: Optional[str] = None,
        broker: Optional[Broker] = None,
    ) -> Optional[KR]:
        """Finds object by name.

        Args:
            name: name of the object
            ns: namespace
            broker: Broker to query K8s API

        Returns:
            return object or None if not found
        """
        if broker is None:
            broker = Broker.get_default_instance()
        raw_res = cls.find_raw(name, ns, broker)
        return cls(raw_res, broker=broker) if raw_res else None

    @classmethod
    def apply(
        cls: Type[KR],
        name: str,
        body: dict[str, Any],
        ns: str,
        broker: Optional[Broker] = None,
    ) -> Any:
        """
        Server-side apply the object in body.
        :param name:
        :param body:
        :param ns:
        :param broker:
        :return:
        """
        if broker is None:
            broker = Broker.get_default_instance()

        fun = cls.k8s_apply_method
        return (
            fun(
                name=name,
                body=body,
                namespace=ns,
                broker=broker,
                force=True,
                field_manager="k8kat",
            )
            if cls.is_namespaced()
            else fun(
                name=name, body=body, broker=broker, force=True, field_manager="k8kat"
            )
        )

    @classmethod
    def delete_if_exists(
        cls,
        name: str,
        ns: str,
        wait_until_gone: Optional[bool] = False,
        broker: Optional[Broker] = None,
    ) -> None:
        """Deletes object by name.

        Args:
            name: name of the object
            ns: namespace
            wait_until_gone: True if you want to wait for
                             the object to disappear.
            broker: Broker to query K8s API

        Returns:
            return object or None if not found
        """
        instance = cls.find(name, ns, broker)
        if instance:
            instance.delete(wait_until_gone)

    @classmethod
    def wait_until_exists(
        cls,
        name: str,
        ns: Optional[str] = None,
    ) -> Any:
        """Waits for the object and returns it.

        Args:
            name: name of the object
            ns: namespace
            broker: Broker to query K8s API

        Returns:
            return object or None if not found
        """
        res = None
        for _ in range(0, 20):
            res = cls.find(name, ns)
            if res:
                break
            time.sleep(1)
        return res



"""
Copyright 2022 nMachine.io

Requirements:
- e2e cluster, configuration from KUBECONFIG or ~/.kube/config
- applied configuration from k8kat/e2e_tests/cluster-e2e.yaml

"""

import unittest

from k8kat import ConfigMap


class TestConfigMap(unittest.TestCase):
    """Set of e2e tests for ConfigMap resource."""

    def test_list_configmap_from_e2e_namespace(self) -> None:
        """Tests listing pods from 'e2e' namespace"""
        cfms = ConfigMap.list(ns="e2e")
        cfms_filtered = [cfm for cfm in cfms if cfm.name.startswith("map-")]
        self.assertEqual(len(cfms_filtered), 2)

    def test_find_configmap_with_not_found_error(self) -> None:
        """Tests finding not-existing configmap"""
        cfm = ConfigMap.find("config-map-does-not-exist", ns="e2e")
        self.assertIsNone(cfm)

    def test_find_configmap_with_data(self) -> None:
        """Tests finding configmap and get data"""
        cfm = ConfigMap.find("map-raw-data", ns="e2e")
        assert cfm  # mypy requires it (TODO: report/find an issue)
        self.assertEqual(cfm.data, {"key": "value"})
        self.assertEqual(cfm.binary_data, {"bin": "dmFsdWUK"})

    def test_configmap_jdata(self) -> None:
        """Tests setting/getting data from configmap"""
        cfm = ConfigMap.find("map-data", ns="e2e")
        assert cfm
        cfm.jpatch({"k1": "v1"})
        self.assertEqual(cfm.jget(), {"k1": "v1"})

        cfm = ConfigMap.find("map-data", ns="e2e")
        assert cfm
        self.assertEqual(cfm.jget(), {"k1": "v1"})

        cfm.jpatch({"k1": "v2"})
        self.assertEqual(cfm.jget(), {"k1": "v2"})

        cfm.jpatch({"k2": "v3"}, merge=True)
        self.assertEqual(cfm.jget(), {"k1": "v2", "k2": "v3"})

        self.assertEqual(
            cfm.jget(key="non-existent-key", default="default-value"), "default-value"
        )

    def test_configmap_ydata(self) -> None:
        """Tests using yget"""
        cfm = ConfigMap.find("map-data", ns="e2e")
        assert cfm
        self.assertEqual(
            cfm.yget(key="non-existent-key", default="default-value"), "default-value"
        )



"""
Copyright 2022 nMachine.io
"""

from __future__ import annotations

from enum import Enum
from http.client import HTTPResponse
from typing import Any, Dict, List, Optional, Set

from kubernetes.client import ApiException, models  # type: ignore
from kubernetes.stream import stream  # type: ignore

from k8kat.broker import Broker
from k8kat.types import IntelDict, PodMetricsDict
from k8kat.utils import classproperty

from . import BaseResource, pod_utils


class PodStatus(Enum):
    """Enum with pod statuses."""

    PENDING = 0
    POSITIVE = 1
    NEGATIVE = 2


class Pod(BaseResource):
    """Pod resource.

    This class inherits from BaseResource and provide:

    1. static methods to list/find/patch/reload pods using K8s API
    2. object properties, methods to get logs, execute commands etc.
    """

    def __init__(self, raw: models.V1Pod, broker: Broker):
        BaseResource.__init__(self, raw, broker)

    @classmethod
    def k8s_list_method(cls, broker: Broker, **kwargs: Any) -> models.V1PodList:
        ns = None
        if "ns" in kwargs:
            ns = kwargs.get("ns")
            del kwargs["ns"]
        if ns:
            return broker.core_v1_api.list_namespaced_pod(namespace=ns, **kwargs)
        return broker.core_v1_api.list_pod_for_all_namespaces(**kwargs)

    @classmethod
    def k8s_read_method(cls, broker: Broker, **kwargs: Any) -> models.V1Pod:
        ns = kwargs.get("ns")
        del kwargs["ns"]
        return broker.core_v1_api.read_namespaced_pod(namespace=ns, **kwargs)

    @classmethod
    def k8s_patch_method(cls, broker: Broker, **kwargs: Any) -> models.V1Pod:
        return broker.core_v1_api.patch_namespaced_pod(**kwargs)

    @classmethod
    def k8s_delete_method(cls, broker: Broker, **kwargs: Any) -> models.V1Status:
        return broker.core_v1_api.delete_namespaced_pod(**kwargs)

    @classmethod
    def k8s_apply_method(cls, broker: Broker, **kwargs: Any) -> None:
        api = broker.dynamic_api.resources.get(api_version="v1", kind=cls.kind)
        return api.server_side_apply(**kwargs)

    @classproperty
    def kind(self) -> str:
        return "Pod"

    @property
    def labels(self) -> Dict[str, str]:
        base = super().labels
        bad_key = "pod-template-hash"
        return {k: base[k] for k in base.keys() if k != bad_key}

    @property
    def phase(self) -> str:
        """Phase of the pod"""
        return self.raw.status.phase

    @property
    def ip(self) -> Optional[str]:
        """IP of the pod"""
        try:
            return self.raw.status.pod_ip
        except Exception:  # pylint: disable=broad-except
            return None

    @property
    def has_parent(self) -> bool:
        """True if there is a parent object"""
        refs = self.raw.metadata.owner_references
        return refs is not None and len(refs) > 0

    def body(self) -> models.V1Pod:
        """Return raw K8s object

        Returns:
            raw pod (V1Pod)"""
        return self.raw

    def containers(self) -> List[models.V1Container]:
        """Return list of containers

        Returns:
            list of containers"""
        return self.raw.spec.containers

    def container(self, index: Optional[int] = 0) -> models.V1Container:
        """Return container details

        Args:
            index: container to return

        Returns:
            container details"""
        return self.raw.spec.containers[index]

    def image(self, index: Optional[int] = 0) -> str:
        """Return container docker image

        Args:
            index: container to return

        Returns:
            container details"""
        return self.container(index).image

    def rs(self) -> None:  # pylint: disable=invalid-name
        """NotImplementedError"""
        raise NotImplementedError

    def dep(self) -> None:
        """NotImplementedError"""
        raise NotImplementedError

    def ternary_status(self) -> PodStatus:
        """Returns pod status.

        Returns:
            PodStatus"""
        if self._is_old_enough_to_call():
            if self.is_working():
                return PodStatus.POSITIVE
            if self.is_broken():
                return PodStatus.NEGATIVE
        return PodStatus.PENDING

    def _is_old_enough_to_call(self) -> bool:
        # TODO: find better way to check if pod structers are already inited
        return self.seconds_existed() > 2

    def is_running_normally(self) -> bool:
        """Returns true if all containers are in running state"""
        if self.is_running():
            main_states = self.main_container_states()
            runners = self.filter_container_states(main_states, "running")
            return len(main_states) == len(runners)
        return False

    def is_pending_normally(self) -> bool:
        """Returns true if pod is pending"""
        return (
            self.is_pending()
            and not self.is_terminating()
            and not self.is_pending_morbidly()
        )

    def is_running_morbidly(self) -> bool:
        """Returns true if pod is running but some of its container not."""
        return (
            self.is_running()
            and not self.is_terminating()
            and not self.is_running_normally()
        )

    def is_working(self) -> bool:
        """Returns true if pod is working"""
        return self.is_running_normally() or self.has_succeeded()

    def is_broken(self) -> bool:
        """Returns True if pod is not running or failed"""
        return (
            self.is_pending_morbidly()
            or self.is_running_morbidly()
            or self.has_failed()
        )

    def is_terminating(self) -> bool:
        """Returns True if pod is in terminating state"""
        states = self.main_container_states()
        terminators = self.filter_container_states(states, "terminated")
        return len(terminators) > 0

    def has_settled(self) -> bool:
        """Returns True if pod is already settled"""
        return not self.is_pending_normally()

    def did_scheduling_fail(self) -> bool:
        """Returns True if sheduling fail"""
        lifecycle_conditions = self.raw.status.conditions
        return self.cond_has_scheduling_failed(lifecycle_conditions)

    def is_pending_morbidly(self) -> bool:
        """Whether this pod is pending because one or more container is failing to start."""
        if self.is_pending():
            if self.did_scheduling_fail():
                return True

            init_states = self.init_container_states()
            waiting_init = self.filter_container_states(init_states, "waiting")
            if len(self.morbid_pending_reasons(waiting_init)) > 0:
                return True

            main_states = self.main_container_states()
            waiting_main = self.filter_container_states(main_states, "waiting")
            if len(self.morbid_pending_reasons(waiting_main)):
                return True

            return False
        return False

    def is_running(self) -> bool:
        """Whether this pod is in the Running state"""
        return self.raw.status.phase == "Running"

    def is_pending(self) -> bool:
        """Whether this pod is in the Pending state"""
        return self.raw.status.phase == "Pending"

    def has_failed(self) -> bool:
        """Whether this pod is in the Failed state"""
        return self.raw.status.phase == "Failed"

    def has_succeeded(self) -> bool:
        """Whether this pod is in the Succeeded state"""
        return self.raw.status.phase == "Succeeded"

    def has_run(self) -> bool:
        """Whether this pod is in the Failed or Succeeded state"""
        return self.has_failed() or self.has_succeeded()

    def cpu_request(self) -> float:
        """Returns pod's total memory limits in bytes."""
        return self.read_resources_req_or_lim("requests", "cpu")

    def cpu_limit(self) -> float:
        """Returns pod's total memory limits in bytes."""
        return self.read_resources_req_or_lim("limits", "cpu")

    def mem_limit(self) -> float:
        """Returns pod's total memory limits in bytes."""
        return self.read_resources_req_or_lim("limits", "memory")

    def mem_request(self) -> float:
        """Returns pod's total memory requests in bytes."""
        return self.read_resources_req_or_lim("requests", "memory")

    def eph_storage_limit(self) -> float:
        """Returns pod's total ephemeral storage limits in bytes."""
        return self.read_resources_req_or_lim("limits", "ephemeral-storage")

    def eph_storage_request(self) -> float:
        """Returns pod's total ephemeral storage requests in bytes."""
        return self.read_resources_req_or_lim("requests", "ephemeral-storage")

    def read_resources_req_or_lim(self, metric: str, resource: str) -> float:
        """Fetches pod's total resource capacity (either limits or requests)
        for either CPU (cores) or memory (bytes)."""

        def container_level_value(container: models.V1Container) -> float:
            return pod_utils.container_req_or_lim(container, metric, resource) or 0

        containers = self.raw.spec.containers or []
        return sum(list(map(container_level_value, containers)))

    @pod_utils.when_running_normally
    # TODO: cache with TTL will be better here
    # @lru_cache(maxsize=128)
    def load_per_pod_metrics(self) -> Optional[List[PodMetricsDict]]:
        """Loads the appropriate metrics dict from k8s metrics API."""
        try:
            self_metrics = self.broker.custom_api.get_namespaced_custom_object(
                group="metrics.k8s.io",
                version="v1beta1",
                namespace=self.namespace,
                plural="pods",
                name=self.name,
            )
            return [self_metrics] if self_metrics else None
        except ApiException:
            return []

    def running_morbidly_intel(self) -> List[IntelDict]:
        """TODO"""
        return []

    def condition_status_intel(self) -> List[IntelDict]:
        """Return list of invalid conditions"""
        intel = []
        for cond in self.body().status.conditions or []:
            if cond.status == "False":
                intel.append(
                    IntelDict(
                        type="StatusCondition", status=cond.reason, message=cond.message
                    )
                )
        return intel

    def intel(self) -> List[IntelDict]:
        """Return list of IntelDict with errors related to the pod"""
        if self.is_broken():
            return [
                *self.bad_events_intel(),
                *self.running_morbidly_intel(),
                *self.bad_events_intel(),
            ]
        return []

    def main_container_status_summaries(self) -> List[Dict]:
        """Return list of containers with theirs names and statuses"""
        summaries = []

        def extract(cont_status: models.V1ContainerStatus) -> Optional[str]:
            try_states = ["running", "waiting", "terminated"]
            tup = next(
                filter(
                    lambda try_state: getattr(cont_status.state, try_state[1]),
                    enumerate(try_states),
                ),
                None,
            )
            return try_states[tup[0]] if tup else None

        for state in self.body().status.container_statuses:
            summaries.append({"name": state.name, "status": extract(state)})

        return summaries

    def container_status_intel(self) -> List[IntelDict]:
        """Return list of containers in status waiting"""
        bad_states: List[models.V1ContainerState] = self.filter_container_states(
            [*self.init_container_states(), *self.main_container_states()], "waiting"
        )

        intel = []
        for state in bad_states:
            intel.append(
                IntelDict(
                    type="ContainerStatus",
                    message=state.waiting.message,
                    status=state.waiting.reason,
                )
            )

        return intel

    def bad_events_intel(self) -> List[IntelDict]:
        """List of bad events connected with this pod."""
        # bad_events = list(filter(KatEvent.is_failure, self.events()))
        # return list(map(KatEvent.intel_bundle, bad_events))
        return []

    def main_container_states(self) -> List[models.V1ContainerState]:
        """Return list of containers' state"""
        statuses = self.body().status.container_statuses or []
        return [status.state for status in statuses]

    def init_container_states(self) -> List[models.V1ContainerState]:
        """Return list of init containers' state"""
        statuses = self.body().status.init_container_statuses or []
        return [status.state for status in statuses]

    @staticmethod
    def cond_has_scheduling_failed(conditions: List[models.V1PodCondition]) -> bool:
        """Checks conditions if there are invalid sheduling state"""

        def finder(name: str) -> Optional[models.V1PodCondition]:
            return next(filter(lambda c: c.type == name, conditions or []), None)  # type: ignore

        relevant_cond = finder("PodScheduled")
        return bool(relevant_cond and relevant_cond.status == "False")

    @staticmethod
    def morbid_pending_reasons(states: List[models.V1ContainerState]) -> Set[str]:
        """Checks containers states and returns list of reason other
        than ContainerCreating PullingImage, PodInitializing."""
        stated_reasons = {state.waiting.reason for state in states}
        good_reasons = {"ContainerCreating", "PullingImage", "PodInitializing"}
        return stated_reasons - good_reasons

    @staticmethod
    def filter_container_states(
        states: List, _type: str
    ) -> List[models.V1ContainerState]:
        """Filter out an attribute from containers states"""
        return [state for state in states if getattr(state, _type)]

    def raw_logs(self, seconds: Optional[int] = 60) -> Optional[str]:
        """Returns logs from pod as raw output from K8s.

        Args:
            seconds: a relative time in seconds before the current
            time from which to show logs.

        Returns:
            list of log lines
        """
        try:
            return self.broker.core_v1_api.read_namespaced_pod_log(
                namespace=self.namespace, name=self.name, since_seconds=seconds
            )
        except ApiException:
            return None

    def clean_logs(self, seconds: Optional[int] = 60) -> str:
        """Returns logs from pod as string.

        Args:
            seconds: a relative time in seconds before the current
            time from which to show logs.


        Returns:
            list of log lines
        """
        _raw_logs = self.raw_logs(seconds)
        return (_raw_logs or "").strip()

    def log_lines(self, seconds: Optional[int] = 60) -> List[str]:
        """Returns logs from pod as list of logs.

        Args:
            seconds: a relative time in seconds before the current
            time from which to show logs.

        Returns:
            list of log lines

        """
        raw_log_str = self.raw_logs(seconds)
        if raw_log_str:
            return raw_log_str.strip("\n").split("\n")
        return []

    @classmethod
    def consume_runner(
        cls, name: str, ns: str, wait_until_gone: Optional[bool] = False
    ) -> Optional[str]:
        """Returns logs from runner

        Args:
            name: name of pod
            ns: namespace
            wait_until_gone: wait for deleting pod

        Returns:
            logs from pod

        """
        pod = cls.wait_until_exists(name, ns)
        if pod:
            pod.wait_until(pod.has_run)
            logs = None
            if pod.has_succeeded():
                logs = pod.raw_logs()
            pod.delete(wait_until_gone)
            return logs
        return None

    @pod_utils.when_running_normally
    def shell_exec(self, command: str) -> Optional[str]:
        """Executes shell command in pod.

        Args:
            command: command to execute

        Returns:
            output from the command
        """
        fmt_command = pod_utils.coerce_cmd_format(command)
        result = stream(
            self.broker.core_v1_api.connect_get_namespaced_pod_exec,
            self.name,
            self.namespace,
            command=fmt_command,
            stderr=True,
            stdin=False,
            stdout=True,
            tty=False,
        )
        return result.strip() if result else None

    def replace_image(
        self, new_image_name: str, index: Optional[int] = 0
    ) -> models.V1Pod:
        """Replaces container's image.

        Args:
            new_image_name: new image
            index: points container to patch

        Returns:
            Patched object
        """
        self.body().spec.containers[index].image = new_image_name
        return self.patch()

    def wait_until_running(self) -> bool:
        """Waits until pod is in running state.

        Returns:
            True if pod is in running state.
        """
        return self.wait_until(self.is_running)

    def invoke_curl(self, **kwargs: Any) -> HTTPResponse:
        """Executes 'curl' command inside Pod.

        TODO: repace kwargs with named args

        Args:
            headers: optional, list of headers
            body: optional, body for post command
            verb: default GET, http method
            url: url
            path: default /

        Return
            HTTPResponse from the command"""
        fmt_command = pod_utils.build_curl_cmd(**kwargs, with_command=True)
        result = self.shell_exec(fmt_command)
        if result is not None:
            result = pod_utils.parse_response(result)
        return result



"""Defines types use in the library.

Copyright 2022 nMachine.io
"""

from typing import List, Optional

from typing_extensions import TypedDict


class UsageDict(TypedDict):
    """Struct with cpu/memory usage"""

    cpu: Optional[str]
    memory: Optional[str]


class ContainerMetricsDict(TypedDict):
    """Struct with container usage"""

    usage: UsageDict


class PodMetricsDict(TypedDict):
    """Struct with containers usage"""

    containers: List[ContainerMetricsDict]


class NodeMetricsDict(TypedDict):
    """Struct with node metrics"""

    usage: UsageDict


class IntelDict(TypedDict):
    """Struct with intel messages"""

    type: str
    status: str
    message: str



"""
Copyright 2022 nMachine.io

Requirements:
- e2e cluster, configuration from KUBECONFIG or ~/.kube/config
- applied configuration from k8kat/e2e_tests/cluster-e2e.yaml

"""

import time
import unittest

from k8kat import Pod, PodStatus


class TestPod(unittest.TestCase):
    """Set of e2e tests for Pod resource."""

    def test_list_pod_from_all_namespaces(self) -> None:
        """Tests listing pods"""
        pods = Pod.list()
        self.assertTrue(len(pods) > 0)

    def test_list_pod_from_e2e_namespace(self) -> None:
        """Tests listing pods from 'e2e' namespace"""
        pods = Pod.list(ns="e2e")
        self.assertEqual(len(pods), 4)

    def test_find_pod_with_not_found_error(self) -> None:
        """Tests finding not-existing Pod"""
        pod = Pod.find("container-does-not-exist", ns="e2e")
        self.assertIsNone(pod)

    def test_find_pod(self) -> None:
        """Tests finding pod"""
        pod = Pod.find("busybox-with-init-containers", ns="e2e")
        assert pod  # mypy requires it (TODO: report/find an issue)
        self.assertEqual(pod.name, "busybox-with-init-containers")
        self.assertEqual(pod.kind, "Pod")
        self.assertEqual(pod.labels, {"name": "busybox-with-init-containers"})
        self.assertEqual(pod.phase, "Running")
        self.assertRegex(pod.ip, r"^\d+\.\d+\.\d+\.\d+$")  # type: ignore
        self.assertEqual(pod.image(), "busybox")
        self.assertEqual(pod.ternary_status(), PodStatus.POSITIVE)
        self.assertTrue(pod.is_running())
        self.assertTrue(pod.is_running_normally())
        self.assertTrue(pod.is_working())
        self.assertTrue(pod.has_settled())
        self.assertFalse(pod.has_parent)
        self.assertFalse(pod.is_pending_normally())
        self.assertFalse(pod.is_running_morbidly())
        self.assertFalse(pod.is_broken())
        self.assertFalse(pod.is_terminating())
        self.assertFalse(pod.did_scheduling_fail())
        self.assertFalse(pod.is_pending_morbidly())
        self.assertFalse(pod.is_pending())
        self.assertFalse(pod.has_failed())
        self.assertFalse(pod.has_succeeded())
        self.assertFalse(pod.has_run())
        self.assertEqual(pod.cpu_request(), 0.25)
        self.assertEqual(pod.cpu_limit(), 0.5)
        self.assertEqual(pod.mem_limit(), 134217728)
        self.assertEqual(pod.mem_request(), 67108864)
        self.assertEqual(pod.eph_storage_limit(), 3221225472.0)
        self.assertEqual(pod.eph_storage_request(), 2147483648.0)

    def test_get_logs(self) -> None:
        """Tests getting logs"""
        pod = Pod.find("busybox-with-init-containers", ns="e2e")
        assert pod
        logs = pod.log_lines()
        self.assertIn("The app is running!", logs)

    def test_shell_exec(self) -> None:
        """Tests executing command"""
        pod = Pod.find("busybox-with-init-containers", ns="e2e")
        assert pod
        result = pod.shell_exec("whoami")
        self.assertEqual(result, "root")

    def test_replace_image(self) -> None:
        """Tests replacing container image"""
        pod = Pod.find("busybox-to-patch", ns="e2e")
        assert pod
        new_image = "busybox:1.35" if pod.image() == "busybox:1.34" else "busybox:1.34"
        pod = pod.replace_image(new_image)
        assert pod
        self.assertEqual(pod.image(), new_image)

    def test_pod_metrics(self) -> None:
        """Tests querying run-time metrics"""
        pod = Pod.find("busybox-with-init-containers", ns="e2e")
        assert pod

        for _ in range(1, 10):
            metrics = pod.load_per_pod_metrics()
            if metrics is None or len(metrics) == 0:
                time.sleep(3)
                continue

        self.assertEqual(len(metrics), 1)
        metrics = metrics[0]["containers"][0]
        self.assertEqual(metrics["name"], "busybox")
        self.assertEqual(list(metrics["usage"].keys()), ["cpu", "memory"])



"""
Copyright 2022 nMachine.io
"""

import os
import tempfile
import unittest

from k8kat import Broker, ConfigException, Configuration, Pod


class TestBroker(unittest.TestCase):
    """E2E tests for Broker, Configuration"""

    def test_default_broker(self) -> None:
        """Tests creating default broker"""
        pods = Pod.list()
        self.assertTrue(len(pods) > 0)

    def test_configure_broker_with_invalid_kubeconfig(self) -> None:
        """Tests loading configuration from invalid config"""

        with self.assertRaises(ConfigException):
            configuration = Configuration(kubeconfig_path="/tmp/not/found/kubeconfig")
            Broker(configuration)

    def test_configure_broker_with_kubeconfig(self) -> None:
        """Tests loading configuration from kubeconfig"""

        path_kubeconfig = os.environ.get(
            "KUBECONFIG", os.path.expanduser("~/.kube/config")
        )
        temp_kubeconfig = None

        try:
            with tempfile.NamedTemporaryFile(delete=False) as ftemp:

                with open(path_kubeconfig, "br") as fcfg:
                    ftemp.write(fcfg.read())

                ftemp.close()
                temp_kubeconfig = ftemp.name

                configuration = Configuration(kubeconfig_path=temp_kubeconfig)
                broker = Broker(configuration)

                pods = Pod.list(broker=broker)
                self.assertTrue(len(pods) > 0)

        finally:
            if temp_kubeconfig:
                os.unlink(temp_kubeconfig)

    def test_configure_broker_with_bearer_token(self) -> None:
        """Tests passing token, api_server"""

        def readfile(env_name: str) -> str:
            with open(os.environ.get(env_name, env_name), "r", encoding="utf8") as fhn:
                return fhn.read().strip()

        token = readfile("E2E_TOKEN_FILE")
        api_server = readfile("E2E_API_SERVER_FILE")

        configuration = Configuration(token=token, api_server=api_server)
        broker = Broker(configuration)
        pods = Pod.list(broker=broker)
        self.assertTrue(len(pods) > 0)



"""K8kat library to inspect your K8s resources.

Copyright 2022 nMachine.io
"""

from .broker import Broker, ConfigException, Configuration
from .res.configmap import ConfigMap
from .res.ingress import Ingress
from .res.pod import Pod, PodStatus
from .res.secret import Secret
from .res.service_account import ServiceAccount
